package org.rosuda.jrs.experimental;

import java.util.Collection;
import java.util.Queue;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;

import javax.script.ScriptContext;
import javax.script.ScriptException;

import org.rosuda.jrs.BaseScriptEngine;
import org.rosuda.jrs.ScriptEngineEx;

/**
 * Execute script commands asynchronously using a pool of script engines.
 * Eval and call always return null, need to use getFuture to get result.
 * <p>
 * Experimental.
 * Consider using doMC, foreach and %dopar% instead.
 * @author Martyn Tebby
 */
public class AsyncScriptEngine extends BaseScriptEngine {

	private Future<?> future;
	private ExecutorService executor;
	private Queue<ScriptEngineEx> engines;

	/**
	 * Create asynchronous ScriptEngine.
	 * @param engines The pool of ScriptEngines to use.
	 */
	public AsyncScriptEngine(Collection<ScriptEngineEx> engines) {
		this.executor = Executors.newFixedThreadPool(engines.size());
		this.engines = new LinkedBlockingQueue<ScriptEngineEx>(engines);
	}

	/** Gets a Future for the result of the last eval or call. */
	@SuppressWarnings("rawtypes")
	public Future getFuture() {
		return future;
	}

	@Override
	public Object eval(final String script, final ScriptContext context) {
		return submit(new EngineCallable<Object>() {
			public Object call(ScriptEngineEx engine) {
				try {
					return engine.eval(script, context);
				}
				catch(ScriptException se) {
					throw new RuntimeException(se);
				}
			}
		});
	}

	@Override
	public <T> T call(final Class<T> returnType, final Object obj,
			final String function, final Object... args) {
		return submit(new EngineCallable<T>() {
			public T call(ScriptEngineEx engine) {
				try {
					return engine.call(returnType, obj, function, args);
				}
				catch(ScriptException se) {
					throw new RuntimeException(se);
				}
			}
		});
	}

	@Override
	public void close() {
		executor.shutdown();
		for(ScriptEngineEx engine : engines) {
			engine.close();	// should these close ???
		}
	}

	private <T> T submit(final EngineCallable<T> callable) {
		future = executor.submit(new Callable<T>() {
			public T call() {
				ScriptEngineEx engine = engines.remove();
				try {
					return callable.call(engine);
				}
				finally {
					engines.add(engine);
				}		
			}
		});
		return null;
	}

	private interface EngineCallable<V> {
		public V call(ScriptEngineEx engine);
	}

}
